import struct
import math

#Should-be-hidden Data Types:
#	Chunk,SubChunk:
#		tag -  string of length 4
#		length - int, length of data
#		data - string of the data itself

#Simple Data Types:
#	Vector:
#		x,y,z - floats
#	PointList:
#		self.points - list of Vectors
#	TagStrings:
#		tag_strings - list of strings
#	BoundingBox:
#		min,max - Vector

#Lightwave-specific Data Types
#	Polygon:
#		flags - integer. content at 0xFC00
#		indices - list of integers
#	PolygonList:
#		type - string of length 4
#		polygons - list of polygons
#	PolygonTagMapping:
#		type - string of length 4
#		mappings - list of tuples (polygon_index, [int] tag) 
#	VertexMapping:
#		type - string of length 4
#		dimension - the dimension of each mapping data
#		name - string
#		mapping - list of tuples (vertex_index, [tuple of floats in size of dimension] data)
#	DiscontinuousVertexMapping:
#		type - string of length 4
#		dimension - the dimension of each mapping data
#		name - string
#		mapping - list of tuples (vertex_index, polygon_index, [tuple of floats in size of dimension] data)
#	Surface:
#		name - string
#		source - string
#		attributes - unparsed SubChunks
#	Clip:
#		index - integer
#		attributes - unparsed SubChunks
#	Layer:
#		chunks - list of chunks that appeared after the layer chunk
#		number - int
#		flags - int
#		pivot - Vector
#		name - string
#		parent - None or int
#		hidden - Boolean

#	LightWaveFile:
#		layers - list of layers (with all their data)
#		tags - TagStrings
#		surfaces - list of surfaces
#		clips - list of clips

class Chunk:
	def __init__(self, data):
		if len(data) < 8:
			raise Exception("The data isn't big enough")
		self.tag = data[0:4]
		self.length = struct.unpack(">L",data[4:8])[0]
		if self.length % 2 != 0:
			length_to_read = self.length+1
		else:
			length_to_read = self.length
		if len(data) != 8 + length_to_read:
			raise Exception("The data size doesn't match the contents")
		if self.length % 2 != 0:
			if ord(data[-1]) != 0:
				raise Exception("The data pad isn't null")
		self.data = data[8:8+self.length]

class SubChunk:
	def __init__(self, data):
		if len(data) < 6:
			raise Exception("The data isn't big enough")
		self.tag = data[0:4]
		self.length = struct.unpack(">H",data[4:6])[0]
		if self.length % 2 != 0:
			length_to_read = self.length+1
		else:
			length_to_read = self.length
		if len(data) != 6 + length_to_read:
			raise Exception("The data size doesn't match the contents")
		if self.length % 2 != 0:
			if ord(data[-1]) != 0:
				raise Exception("The data pad isn't null")
		self.data = data[6:6+self.length]

def ExtractChunk(data):
	if len(data) < 8:
		raise Exception("The data isn't big enough")
	length = struct.unpack(">L",data[4:8])[0]
	if length%2 != 0:
		length += 1
	length += 8
	return (Chunk(data[:length]), data[length:])

def ExtractSubChunk(data):
	if len(data) < 6:
		raise Exception("The data isn't big enough")
	length = struct.unpack(">H",data[4:6])[0]
	if length%2 != 0:
		length += 1
	length += 6
	return (SubChunk(data[:length]), data[length:])
	
class Vector:
	def __init__(self, data=None):
		if data == None:
			(self.x,self.y,self.z) = (0.0,0.0,0.0)
			return
		if len(data) != 12:
			raise Exception("The data size doesn't match a vector")
		self.x,self.y,self.z = struct.unpack(">fff",data)
	def Cross(self,other):
		result = Vector()
		result.x = self.y*other.z - self.z*other.y
		result.y = self.z*other.x - self.x*other.z
		result.z = self.x*other.y - self.y*other.x
		return result
	def Get(self):
		return (self.x,self.y,self.z)
	def Set(self,value_tuple):
		(self.x,self.y,self.z) = value_tuple
	def __add__(self,other):
		new = Vector()
		new.Set(map(float.__add__,self.Get(),other.Get()))
		return new
	def __sub__(self,other):
		new = Vector()
		new.Set(map(float.__sub__,self.Get(),other.Get()))
		return new
	def IsZero(self):
		return self.x == 0 and self.y == 0 and self.z == 0
	def Normalized(self):
		length2 = self.x*self.x + self.y*self.y + self.z*self.z
		length = math.sqrt(length2)
		new = Vector()
		new.Set(map(float.__div__,self.Get(),[length]*3))
		return new

def ExtractString(data):
	null_index = data.index(chr(0))
	string = data[0:null_index]
	if null_index % 2 == 0:
		null_index += 1
		if data[null_index] != chr(0):
			raise Exception("The string pad isn't null")
	return (string,data[null_index+1:])

class PointList:
	def __init__(self, chunk):
		if chunk.tag != "PNTS":
			raise Exception("The chunk provided isn't a point list")
		point_count = chunk.length / 12
		if chunk.length != point_count * 12:
			raise Exception("The data size cannot contain points")
		self.points = []
		for i in xrange(point_count):
			self.points.append(Vector(chunk.data[i*12:i*12+12]))

def ParseVariableIndex(data):
	if data[0] == chr(0xFF):
		index = struct.unpack(">L",data[0:4])[0]
		return index&0x00FFFFFF, data[4:]
	else:
		index = struct.unpack(">H",data[0:2])[0]
		return index, data[2:]

class Polygon:
	def __init__(self, flags, indices):
		self.flags = flags
		self.indices = indices

def ExtractPolygon(data):
	#first two bytes signify flags and the number of indices
	numvert_flags = struct.unpack(">H", data[0:2])[0]
	vertex_count = numvert_flags & 0x03FF
	flags = numvert_flags & 0xFC00
	data = data[2:]
	
	#the rest is (vertex_count) variable-sized indices
	indices = []
	for i in xrange(vertex_count):
		index,data = ParseVariableIndex(data)
		indices.append(index)
	return Polygon(flags,indices), data

class PolygonList:
	def __init__(self, chunk):
		if chunk.tag != "POLS":
			raise Exception("The chunk provided isn't a polygon list")
		
		if chunk.length < 4:
			raise Exception("The data isn't big enough")
		self.type = chunk.data[0:4]
		self.polygons = []
		toparse = chunk.data[4:]
		while len(toparse) != 0:
			poly,toparse = ExtractPolygon(toparse)
			self.polygons.append(poly)

class BoundingBox:
	def __init__(self, chunk):
		if chunk.tag != "BBOX":
			raise Exception("The chunk provided isn't a polygon list")
		
		if chunk.length != 24:
			raise Exception("The data size doesn't match the contents")
		self.min = Vector(chunk.data[0:12])
		self.max = Vector(chunk.data[12:24])

class PolygonTagMapping:
	def __init__(self, chunk):
		if chunk.tag != "PTAG":
			raise Exception("The chunk provided isn't a polygon tag mapping")
		
		if chunk.length < 4:
			raise Exception("The data isn't big enough")
		self.type = chunk.data[0:4]
		data = chunk.data[4:]
		
		self.mappings = [] #(poly,tag)*
		while len(data) != 0:
			poly, data = ParseVariableIndex(data)
			tag = struct.unpack(">H",data[0:2])[0]
			self.mappings.append((poly,tag))
			data = data[2:]

class VertexMapping:
	def __init__(self, chunk):
		if chunk.tag != "VMAP":
			raise Exception("The chunk provided isn't a vertex mapping")
		
		if chunk.length < 6:
			raise Exception("The data isn't big enough")
		self.type = chunk.data[0:4]
		self.dimension = struct.unpack(">H",chunk.data[4:6])[0]
		self.name,data = ExtractString(chunk.data[6:])
		
		self.mapping = [] #(vert,value)*
		while len(data) != 0:
			vert, data = ParseVariableIndex(data)
			value = []
			for i in range(self.dimension):
				value.append(struct.unpack(">f",data[4*i:4*i+4])[0])
			self.mapping.append((vert,tuple(value)))
			data = data[4*self.dimension:]

class DiscontinuousVertexMapping:
	def __init__(self, chunk):
		if chunk.tag != "VMAD":
			raise Exception("The chunk provided isn't a discontinuous vertex mapping")
		
		if chunk.length < 6:
			raise Exception("The data isn't big enough")
		self.type = chunk.data[0:4]
		self.dimension = struct.unpack(">H",chunk.data[4:6])[0]
		self.name,data = ExtractString(chunk.data[6:])
		
		self.mapping = [] #(vert,value)*
		while len(data) != 0:
			vert, data = ParseVariableIndex(data)
			poly, data = ParseVariableIndex(data)
			value = []
			for i in range(self.dimension):
				value.append(struct.unpack(">f",data[4*i:4*i+4])[0])
			self.mapping.append((vert,poly,tuple(value)))
			data = data[4*self.dimension:]

class Surface:
	def __init__(self, chunk):
		if chunk.tag != "SURF":
			raise Exception("The chunk provided isn't a surface")
		
		data = chunk.data
		self.name,data = ExtractString(data)
		self.source,data = ExtractString(data)
		self.attributes = []
		while len(data) != 0:
			subchunk,data = ExtractSubChunk(data)
			self.attributes.append(subchunk)

class Clip:
	def __init__(self, chunk):
		if chunk.tag != "CLIP":
			raise Exception("The chunk provided isn't a clip")
		
		self.index = struct.unpack(">L",chunk.data[0:4])[0]
		data = chunk.data[4:]
		self.attributes = []
		while len(data) != 0:
			subchunk,data = ExtractSubChunk(data)
			self.attributes.append(subchunk)

class TagStrings:
	def __init__(self, chunk):
		if chunk.tag != "TAGS":
			raise Exception("The chunk provided isn't a tag strings")
		
		data = chunk.data
		self.tag_strings = []
		while len(data) != 0:
			string,data = ExtractString(data)
			self.tag_strings.append(string)

class Layer:
	def __init__(self, chunk):
		self.chunks = []
		
		if chunk.tag != "LAYR":
			raise Exception("The chunk provided isn't a layer")
		if chunk.length < 16:
			raise Exception("The data isn't big enough")
		self.number = struct.unpack(">H",chunk.data[0:2])[0]
		self.flags = struct.unpack(">H",chunk.data[2:4])[0]
		self.pivot = Vector(chunk.data[4:16])
		self.name,remainder = ExtractString(chunk.data[16:])
		
		#the remainder (if there is) specifies the parent
		if len(remainder) == 0:
			self.parent = None
		elif len(remainder) == 2:
			self.parent = struct.unpack(">H",remainder)[0]
			if self.parent == -1:
				self.parent == None
		else:
			raise Exception("The data size doesn't match Layer format")
		
		#MSB in flags signifies whether the layer is hidden
		self.hidden = False
		if self.flags & 0x8000:
			self.hidden = True
	
	def append(self, chunk):
		if chunk.tag == "PNTS":
			chunk = PointList(chunk)
		elif chunk.tag == "POLS":
			chunk = PolygonList(chunk)
		elif chunk.tag == "BBOX":
			chunk = BoundingBox(chunk)
		elif chunk.tag == "PTAG":
			chunk = PolygonTagMapping(chunk)
		elif chunk.tag == "VMAP":
			chunk = VertexMapping(chunk)
		elif chunk.tag == "VMAD":
			chunk = DiscontinuousVertexMapping(chunk)
		else:
			print "%s added to layer %s" % (chunk.tag,self.name)
		self.chunks.append(chunk)
	
class LightWaveFile:
	def __init__(self, filename):
		#a file is a Chunk of type FORM
		contents = file(filename,"rb").read()
		datachunk = Chunk(contents)
		if datachunk.tag != "FORM":
			raise Exception("File format error")
		data = datachunk.data
		
		#that chunk contains its type LWO2
		if data[0:4] != "LWO2":
			raise Exception("File format error")
		data = data[4:]
		
		#and then a list of chunks
		chunk_list = []
		while len(data) != 0:
			chunk,remainder = ExtractChunk(data)
			chunk_list.append(chunk)
			data = remainder
		
		self.layers = []
		self.tags = None #only one
		self.surfaces = []
		self.clips = []
		add_to = None
		for chunk in chunk_list:
			if chunk.tag == "LAYR":
				layer = Layer(chunk)
				add_to = layer
				self.layers.append(layer)
			elif chunk.tag == "TAGS":
				if self.tags != None:
					raise Exception("Tag strings chunk appeared twice")
				self.tags = TagStrings(chunk)
			elif chunk.tag == "SURF":
				self.surfaces.append(Surface(chunk))
			elif chunk.tag == "CLIP":
				self.clips.append(Clip(chunk))
			else:
				if add_to == None:
					raise Exception("Unrecognized chunk %s before layer", chunk.tag)
				add_to.append(chunk)
