import sys

import LightwaveParser
from LightwaveParser import Vector

SMOOTH_LAYERS = False

class PointData:
	def __init__(self):
		self.v = None
		self.vt = None
		self.vn = None

class FaceData:
	def __init__(self):
		self.normal = (0,0,0)
		self.points = []
	
	def CalculateNormal(self,vertex_array):
		def vertex_index(face_point_index):
			return self.points[face_point_index%degree].v
		degree = len(self.points)
		if degree < 3:
			raise Exception("Polygon must have at least 3 vertices")
		normal = Vector()
		index = 0
		while index < len(self.points):
			v1 = vertex_array.points[vertex_index(index+1)] - vertex_array.points[vertex_index(index+0)]
			v2 = vertex_array.points[vertex_index(index+2)] - vertex_array.points[vertex_index(index+1)]
			normal += v1.Cross(v2)
			index += 1
		if normal.IsZero():
			raise Exception("Normals sum for the face is zero")
		return normal.Normalized()
	
	def DumpOutput(self):
		face_datastring = []
		for point_data in self.points:
			v = str(point_data.v)
			vt = ""
			if point_data.vt != None:
				vt = str(point_data.vt)
			vn = ""
			if point_data.vn != None:
				vn = str(point_data.vn)
			else:
				print self
			face_datastring.append("/".join((v,vt,vn)))
		return "f %s\n" % dump_list(face_datastring)

def dump_list(list):
	return " ".join(map(str,list))

def extract_instance(list, type, cond):
	for item in list:
		if isinstance(item, type) and cond(item):
			list.remove(item)
			return item
	return None

def parse_layer_texture(layer):
	output_data = ""
	
	#create and dump a list of texture coordinates
	texture_coord_list = set()
	texture_coord_list.add((0.0,0.0))
	for map_info in layer.texture_map.mapping:
		texture_coord_list.add(map_info[1])
	if layer.d_texture_map != None:
		for map_info in layer.d_texture_map.mapping:
			texture_coord_list.add(map_info[2])
	texture_coord_list = list(texture_coord_list)
	for tex_coord in texture_coord_list:
		output_data += "vt %s\n" % dump_list(tex_coord)
	
	#create real texture coordinates mapping (real = dict)
	vertex_texture_mapping = dict( layer.texture_map.mapping )
	if len(vertex_texture_mapping) != len(layer.texture_map.mapping):
		raise Exception("Repetition have been found at the texture map mappings")
	if layer.d_texture_map != None:
		face_vertex_texture_mapping = dict(map(lambda t: ((t[0],t[1]),t[2]), layer.d_texture_map.mapping ))
		if len(face_vertex_texture_mapping) != len(layer.d_texture_map.mapping):
			raise Exception("Repetition have been found at the texture map discontinuous mappings")
	
	#create texture information for each vertex in polygon
	for face_index in xrange(len(layer.faces_data)):
		face_data = layer.faces_data[face_index]			
		for point_data in face_data.points:
			vertex_index = point_data.v
			if vertex_index in vertex_texture_mapping.keys():
				vertex_texture = vertex_texture_mapping[vertex_index]
			else:
				vertex_texture = (0.0,0.0)
			if layer.d_texture_map != None:
				if (vertex_index,face_index) in face_vertex_texture_mapping:
					vertex_texture = face_vertex_texture_mapping[(vertex_index,face_index)]
			point_data.vt = texture_coord_list.index(vertex_texture)
	
	return output_data

def dump_layer_information(layer,parsed_file):
	output_data = ""
	chunks = layer.chunks[:] #shallow copy
	
	#parse the point list and output it
	layer.point_list = extract_instance(chunks, LightwaveParser.PointList,
		lambda chunk:True)
	output_data += "\n#Vertex list\n\n"
	for p in layer.point_list.points:
		output_data += "v %s\n" % dump_list(p.Get())
	
	#parse and process the face list
	face_list = extract_instance(chunks, LightwaveParser.PolygonList,
		lambda chunk: chunk.type == 'FACE')
	layer.faces_data = []
	for p in face_list.polygons:
		if p.flags != 0:
			raise Exception("Unhandled polygon flags")
		face_data = FaceData()
		for vertex_index in p.indices:
			new_point = PointData()
			new_point.v = vertex_index
			face_data.points.append(new_point)
		layer.faces_data.append(face_data)
	
	#parse the texture map
	layer.texture_map = extract_instance(chunks, LightwaveParser.VertexMapping,
		lambda chunk:chunk.type == 'TXUV')
	layer.d_texture_map = extract_instance(chunks, LightwaveParser.DiscontinuousVertexMapping,
		lambda chunk:chunk.type == 'TXUV')
	#if there is a texture, process it
	output_data += "\n#Texture coordinates list\n\n"
	if layer.texture_map != None:
		output_data += parse_layer_texture(layer)
	
	#create a normal for each face
	for face_data in layer.faces_data:
		face_data.normal = face_data.CalculateNormal(layer.point_list)
	
	#extract the surface list and sort the polygons into surface-group
	surfaces = {}
	layer.surface_mapping = extract_instance(chunks, LightwaveParser.PolygonTagMapping,
		lambda chunk:chunk.type == 'SURF')
	for mapping in layer.surface_mapping.mappings:
		if not surfaces.has_key(mapping[1]):
			surfaces[mapping[1]] = []
		surfaces[mapping[1]].append(mapping[0])
	
	layer.normal_list = []
	
	if SMOOTH_LAYERS:
		#calculate normals per-surface
		for surface in surfaces:
			#initialize the normals of each vertex (per surface)
			surface_vertex_normals = []
			for point in layer.point_list.points:
				surface_vertex_normals.append(Vector())
			#sum the normals for each point
			for face_index in surfaces[surface]:
				face_data = layer.faces_data[face_index]
				for point_data in face_data.points:
					surface_vertex_normals[point_data.v] += face_data.normal
			#reNormalize these normals
			for normal in surface_vertex_normals:
				if not normal.IsZero():
					normal.Set(normal.Normalized().Get())
			
			#set the face data to point to the proper normals
			for face_index in surfaces[surface]:
				face_data = layer.faces_data[face_index]
				for point_data in face_data.points:
					point_normal = surface_vertex_normals[point_data.v]
					normal_raw = point_normal.Get()
					if normal_raw not in layer.normal_list:
						layer.normal_list.append(normal_raw)
					point_data.vn = layer.normal_list.index(normal_raw)
	else:
		#calculate normals per-surface
		for surface in surfaces:
			#set the face data to point to the proper normals
			for face_index in surfaces[surface]:
				face_data = layer.faces_data[face_index]
				for point_data in face_data.points:
					normal_raw = face_data.normal.Get()
					if normal_raw not in layer.normal_list:
						layer.normal_list.append(normal_raw)
					point_data.vn = layer.normal_list.index(normal_raw)
	
	#dump the normals
	output_data += "\n#Normal list\n\n"
	for normal in layer.normal_list:
		output_data += "vn %s\n" % dump_list(normal)
	
	#dump the faces list (indexed of: vertices, texture coordinates, and normals) into the output buffer
	output_data += "\n#Face list\n\n"
	for surface in surfaces:
		output_data += "#%s:\n" % parsed_file.tags.tag_strings[surface]
		for face_index in surfaces[surface]:
			face_data = layer.faces_data[face_index]
			output_data += face_data.DumpOutput()
	
	#check out all the other chunks of the layer
	for chunk in chunks:
		if isinstance(chunk,LightwaveParser.PolygonTagMapping) and chunk.type == 'COLR':
			pass #polygon - color mapping for viewing sketch
		elif isinstance(chunk,LightwaveParser.BoundingBox):
			pass #Bounding box of a layer
		else:
			print "  Unhandled chunk:",chunk,chunk.type
	return output_data

def main(filename):
	global parsed_file
	parsed_file = LightwaveParser.LightWaveFile(filename)
	print "Tags:",parsed_file.tags.tag_strings
	for i in xrange(len(parsed_file.layers)):
		layer = parsed_file.layers[i]
		if layer.name == "":
			continue
		print "%d: Layer %d, named \"%s\"" % (i, layer.number, layer.name)
		output_data = dump_layer_information(layer,parsed_file)
		file("%s.obj" % layer.name,"w").write(output_data)

if __name__ == "__main__":
	if len(sys.argv) == 2:
		main(sys.argv[1])
	else:
		print "Usage: %s <lwo filename>" % (sys.argv[0])