import Image
from OpenGL import GL
from pgu import gui
import pygame

class Sprite:
	def __init__(self, image):
		self.size = image.size
		
		self.texture = GL.glGenTextures(1)
		GL.glBindTexture(GL.GL_TEXTURE_2D, self.texture)
		GL.glTexImage2D(
			GL.GL_TEXTURE_2D, #target
			0, #level
			4, #internal format
			image.size[0], image.size[1], #size
			0, #border
			GL.GL_RGBA, #format
			GL.GL_UNSIGNED_BYTE,
			image.tostring())
		GL.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MAG_FILTER, GL.GL_NEAREST)
		GL.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MIN_FILTER, GL.GL_NEAREST)
		GL.glBindTexture(GL.GL_TEXTURE_2D, 0)
	
	def draw(self):
		GL.glBindTexture(GL.GL_TEXTURE_2D, self.texture)
		GL.glPushMatrix()
		GL.glScalef(self.size[0],self.size[1],1)
		GL.glBegin(GL.GL_QUADS)
		for pair in [(0,0),(0,1),(1,1),(1,0)]:
			GL.glTexCoord2f(*pair)
			GL.glVertex2f(*pair)
		GL.glEnd()
		GL.glPopMatrix()
		
		GL.glFinish()
		GL.glBindTexture(GL.GL_TEXTURE_2D, 0)
	
	def __del__(self):
		GL.glDeleteTextures([self.texture])

class TextSprite(Sprite):
	def __init__(self, image, data):
		Sprite.__init__(self, image)
		
		config = {}
		for char_index in xrange(len(data[0])):
			start_pos = data[1][char_index]
			width = data[1][char_index+1] - data[1][char_index]
			config[ data[0][char_index] ] = (start_pos,width)
		self.config = config
	
	def draw(self, text):
		GL.glBindTexture(GL.GL_TEXTURE_2D, self.texture)
		
		position = 0
		for char in text:
			start_length = self.config[char][0]
			width = self.config[char][1]
			
			GL.glMatrixMode(GL.GL_TEXTURE)
			GL.glPushMatrix()
			GL.glTranslatef(start_length/float(self.size[0]),0,0)
			GL.glScalef(width/float(self.size[0]),1,1)
			
			GL.glMatrixMode(GL.GL_MODELVIEW)
			GL.glPushMatrix()
			GL.glTranslatef(position,0,0)
			GL.glScalef(width,self.size[1],1)
			position += width
			
			GL.glBegin(GL.GL_QUADS)
			for pair in [(0,0),(0,1),(1,1),(1,0)]:
				GL.glTexCoord2f(*pair)
				GL.glVertex2f(*pair)
			GL.glEnd()
			
			GL.glMatrixMode(GL.GL_TEXTURE)
			GL.glPopMatrix()
			GL.glMatrixMode(GL.GL_MODELVIEW)
			GL.glPopMatrix()
		
		GL.glFinish()
		GL.glBindTexture(GL.GL_TEXTURE_2D, 0)

class GUISprite:
	@staticmethod
	def Get2Expo(number):
		retval = 1
		while retval<number:
			retval *= 2
		return retval
	
	def __init__(self, width, height, theme = None):
		self.size = (width, height)
		self.texture_size = (GUISprite.Get2Expo(width),GUISprite.Get2Expo(height))
		
		self.texture = GL.glGenTextures(1)
		self.need_update = True
		GL.glBindTexture(GL.GL_TEXTURE_2D, self.texture)
		GL.glTexParameterf(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MAG_FILTER, GL.GL_NEAREST)
		GL.glTexParameterf(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MIN_FILTER, GL.GL_NEAREST)
		GL.glBindTexture(GL.GL_TEXTURE_2D, 0)
		
		self.gui = gui.App(theme = theme)
		self.container = gui.Container(width=800, height=600)
		
		self.surface = pygame.Surface(
			(self.texture_size[0], self.texture_size[1]),
			pygame.SRCALPHA, 32)
		self.surface = self.surface.convert_alpha()
		self.usable = (
			float(self.size[0]) / self.texture_size[0],
			float(self.size[1]) / self.texture_size[1] )
		self.guiSurface = self.surface.subsurface(
			0, 0,
			self.size[0], self.size[1])
		
		self.gui.init(self.container, self.guiSurface)
	
	def add(self, widget, x, y):
		self.container.add(widget, x, y)
		self.gui.init(self.container, self.guiSurface)
		self.need_update = True
	
	def remove(self, widget):
		self.container.remove(widget)
		self.gui.init(self.container, self.guiSurface)
		self.need_update = True
	
	def update(self):
		self.gui.init(self.container, self.guiSurface)
		self.need_update = True
	
	def handle_event(self, event):
		self.gui.event(event)
		self.need_update = True
	
	def update_texture(self):
		self.need_update = False
		self.gui.update(self.guiSurface)
		textureData = pygame.image.tostring(self.surface, "RGBA")
		
		GL.glClear(GL.GL_COLOR_BUFFER_BIT)
		GL.glClear(GL.GL_COLOR_BUFFER_BIT)
		GL.glClear(GL.GL_COLOR_BUFFER_BIT)
		GL.glTexImage2D(
			GL.GL_TEXTURE_2D, 
			0,
			4,
			self.texture_size[0], self.texture_size[1],
			0,
			GL.GL_RGBA,
			GL.GL_UNSIGNED_BYTE,
			textureData)
	
	def draw(self):
		GL.glBindTexture(GL.GL_TEXTURE_2D, self.texture)
		
		if self.need_update:
			self.update_texture()
		
		#draw
		GL.glBegin(GL.GL_QUADS)	
		GL.glTexCoord2f(0,				0)
		GL.glVertex2f(	0,				0)
		GL.glTexCoord2f(0,				self.usable[1])
		GL.glVertex2f(	0,				self.size[1])
		GL.glTexCoord2f(self.usable[0],	self.usable[1])
		GL.glVertex2f(	self.size[0],	self.size[1])
		GL.glTexCoord2f(self.usable[0],	0)
		GL.glVertex2f(	self.size[0],	0)
		GL.glEnd()
		
		GL.glBindTexture(GL.GL_TEXTURE_2D, 0)
	
	def __del__(self):
		GL.glDeleteTextures([self.texture])